import logging
from enum import Enum
from math import ceil, floor


logger = logging.getLogger(__name__)


class RoundingScheme(Enum):
    """
    Defines the kind of rounding method.
    """
    floor = "floor"
    ceil = "ceil"
    near = "Near"


def round_to_int(
    value: float, scheme: RoundingScheme = RoundingScheme.floor
) -> int:
    """
    Rounds a floating-point value to an integer, using a given rounding method.

    Parameters
    ----------
    value : float
        the floating-point value to round
    scheme : RoundingScheme
        the method of rounding, defaults to floor
    """
    if scheme == RoundingScheme.floor:
        return floor(value)
    elif scheme == RoundingScheme.ceil:
        return ceil(value)
    elif scheme == RoundingScheme.near:
        return round(value)
    else:
        logger.warning(f'tried rounding with a invalid scheme ({scheme}) using floor instead')
        return int(floor(value))



from enum import Enum


class DataTypes(Enum):
    """Defines the data point data types supported by the SGr specification."""

    INT = "INT"
    FLOAT = "FLOAT"
    STRING = "STRING"
    ENUM = "ENUM"
    BOOLEAN = "BOOLEAN"
    BITMAP = "BITMAP"
    DATE_TIME = "DATE_TIME"
    JSON = "JSON"



from datetime import datetime
from typing import Any

from sgr_specification.v0.generic import EnumMapProduct

from sgr_commhandler.api import DataPointValidator
from sgr_commhandler.api.data_types import DataTypes

INT_SIZES: set[int] = {8, 16, 32, 64}
FLOAT_SIZES: set[int] = {32, 64}


class UnsupportedValidator(DataPointValidator):
    """
    Implements a default validator for all unsupported data types.
    """
    
    def validate(self, value: Any) -> bool:
        return False


class EnumValidator(DataPointValidator):
    """
    Implements a validator for enumeration values.
    """

    def __init__(self, type: EnumMapProduct):
        if type and type.enum_entry:
            self._valid_ordinals: set[int] = {
                entry.ordinal
                for entry in type.enum_entry
                if entry.ordinal is not None
            }
            self._valid_literals: set[str] = {
                entry.literal
                for entry in type.enum_entry
                if entry.literal is not None
            }
            self._options: list[tuple[str, int]] = [
                (entry.literal, entry.ordinal)
                for entry in type.enum_entry
                if entry.ordinal is not None and entry.literal is not None
            ]
        else:
            self._valid_literals: set[str] = set()
            self._valid_ordinals: set[int] = set()
            self._options: list[tuple[str, int]] = []

    def validate(self, value: Any) -> bool:
        if value is None:
            return False
        return (isinstance(value, str) and value in self._valid_literals) or (
            isinstance(value, int) and value in self._valid_ordinals
        )

    def data_type(self) -> DataTypes:
        return DataTypes.ENUM

    def options(self) -> list[Any]:
        return self._options


class IntValidator(DataPointValidator):
    """
    Implements a validator for integer values.
    """

    def __init__(self, size: int, signed: bool = True):
        self._size = size if size in INT_SIZES else next(iter(INT_SIZES))
        if signed:
            self._lower_bound = -(2 ** (self._size - 1))
            self._upper_bound = (2 ** (self._size - 1)) - 1
        else:
            # TODO verify that this even works with uint64
            self._lower_bound = 0
            self._upper_bound = (2**self._size) - 1

    def validate(self, value: Any) -> bool:
        if value is None:
            return False
        if isinstance(value, float):
            return self._lower_bound <= value <= self._upper_bound
        try:
            return self._lower_bound <= int(value) <= self._upper_bound
        except Exception:
            return False

    def data_type(self) -> DataTypes:
        return DataTypes.INT


class FloatValidator(DataPointValidator):
    """
    Implements a validator for floating-point values.
    """

    def __init__(self, size: int):
        self._size = size if size in FLOAT_SIZES else next(iter(FLOAT_SIZES))

    def validate(self, value: Any) -> bool:
        if value is None:
            return False
        if isinstance(value, float):
            return True
        try:
            float(value)
            return True
        except Exception:
            return False

    def data_type(self) -> DataTypes:
        return DataTypes.FLOAT


class StringValidator(DataPointValidator):
    """
    Implements a validator for text strings.
    """

    def validate(self, value: Any) -> bool:
        if value is None:
            return False
        if isinstance(value, str):
            return True
        try:
            str(value)
            return True
        except Exception:
            return False

    def data_type(self) -> DataTypes:
        return DataTypes.STRING


class BooleanValidator(DataPointValidator):
    """
    Implements a validator for boolean values.
    """

    def validate(self, value: Any) -> bool:
        if value is None:
            return False
        if isinstance(value, bool):
            return True
        try:
            bool(value)
            return True
        except Exception:
            return False

    def data_type(self) -> DataTypes:
        return DataTypes.BOOLEAN


class BitmapValidator(DataPointValidator):
    """
    Implements a validator for bitmap values.
    """

    def validate(self, value: Any) -> bool:
        if value is None:
            return False
        return isinstance(value, dict)

    def data_type(self) -> DataTypes:
        return DataTypes.BITMAP


class DateTimeValidator(DataPointValidator):
    """
    Implements a validator for date/time values.
    """

    def validate(self, value: Any) -> bool:
        if value is None:
            return False
        # TODO fully implement conversion
        return isinstance(value, datetime)

    def data_type(self) -> DataTypes:
        return DataTypes.DATE_TIME


class JsonValidator(DataPointValidator):
    """
    Implements a validator for JSON values.
    """

    def validate(self, value: Any) -> bool:
        if value is None:
            return False
        return isinstance(value, dict)

    def data_type(self) -> DataTypes:
        return DataTypes.JSON



import json
import logging
from typing import Awaitable, Callable, TypeAlias

import aiohttp
import jmespath
from aiohttp.client import ClientSession
from jmespath.exceptions import JMESPathError
from sgr_specification.v0.product import RestApiInterface
from sgr_specification.v0.product.rest_api_types import (
    HeaderList,
    RestApiAuthenticationMethod,
)

logger = logging.getLogger(__name__)

Authenticator: TypeAlias = Callable[
    [RestApiInterface, ClientSession], Awaitable[bool]
]


async def authenticate_not(
    interface: RestApiInterface, session: ClientSession
) -> bool:
    """
    Skips authentication.

    Parameters
    ----------
    interface : RestApiInterface
        the device interface
    session : ClientSession
        the REST client session
    
    Returns
    -------
    bool
        True if authenticated, False otherwise
    """

    return True


async def authenticate_with_bearer_token(
    interface: RestApiInterface, session: ClientSession
) -> bool:
    """
    Authenticates using Bearer token (JWT).

    Parameters
    ----------
    interface : RestApiInterface
        the device interface
    session : ClientSession
        the REST client session
    
    Returns
    -------
    bool
        True if authenticated, False otherwise
    """

    try:
        description = interface.rest_api_interface_description
        if description is None:
            raise Exception("no interface description")
        base_url = description.rest_api_uri
        if base_url is None:
            raise Exception("no base URL")
        bearer_option = description.rest_api_bearer
        if bearer_option is None:
            raise Exception("no Bearer option")
        service_call = bearer_option.rest_api_service_call
        if service_call is None:
            raise Exception("no REST service call for authentication")
        request_path = service_call.request_path
        if request_path is None:
            raise Exception("no request path")

        authentication_url = f"{base_url}{request_path}"
        logger.debug("auth URL = " + authentication_url)

        headers = {
            header_entry.header_name: header_entry.value
            for header_entry in (
                service_call.request_header
                if service_call.request_header
                else HeaderList()
            ).header
        }

        request_body = service_call.request_body
        if request_body is None:
            raise Exception("no request body")

        data = json.loads(request_body)
        async with session.post(
            url=authentication_url,
            headers=headers,
            json=data,
        ) as res:
            if 200 <= res.status < 300:
                logger.info(f"Authentication successful: Status {res.status}")
                try:
                    response = await res.text()
                    token = jmespath.search("accessToken", json.loads(response))
                    if token:
                        session.headers.update(
                            {"Authorization": f"Bearer {token}"}
                        )

                        logger.info("Token retrieved successfully")
                        return True
                    else:
                        logger.warning("Token not found in the response")
                        return False
                except json.JSONDecodeError:
                    logger.error("Failed to decode JSON response")
                except JMESPathError:
                    logger.error("Failed to search JSON data using JMESPath")
            else:
                logger.warning(f"Authentication failed: Status {res.status}")
                logger.info(f"Response: {await res.text()}")

    except aiohttp.ClientError as e:
        logger.error(f"Network error occurred: {e}")
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
    finally:
        return False


supported_authentication_methods: dict[
    RestApiAuthenticationMethod, Authenticator
] = {
    RestApiAuthenticationMethod.NO_SECURITY_SCHEME: authenticate_not,
    RestApiAuthenticationMethod.BEARER_SECURITY_SCHEME: authenticate_with_bearer_token,
}


async def setup_authentication(
    rest_api_interface: RestApiInterface, session: ClientSession
) -> bool:
    """
    Performs authentication asynchronously, depending on the configured method.

    Parameters
    ----------
    interface : RestApiInterface
        the device interface
    session : ClientSession
        the REST client session
    
    Returns
    -------
    bool
        True if authenticated, False otherwise
    """
    descr = rest_api_interface.rest_api_interface_description
    if descr is None:
        raise Exception("no REST interface description")
    method = descr.rest_api_authentication_method
    if method is None:
        raise Exception("no authentication method")
    auth_fn = supported_authentication_methods.get(method)
    if auth_fn is None:
        raise Exception("unsupported authentication method")
    return await auth_fn(rest_api_interface, session)



import asyncio
import logging
from threading import Lock

from sgr_specification.v0.product.modbus_types import BitOrder

from sgr_commhandler.driver.modbus.modbus_client_async import (
    SGrModbusClient,
    SGrModbusRTUClient,
)

logger = logging.getLogger(__name__)


class ModbusClientWrapper:
    """
    Implements a Modbus client wrapper for shared use.
    """

    def __init__(
        self, identifier: str, client: SGrModbusClient, shared: bool = False
    ):
        self.identifier = identifier
        self.client = client
        self.shared = shared
        self.registered_devices = set()
        self.connected_devices = set()

    async def connect(self, device_id: str):
        if self.shared:
            if device_id not in self.registered_devices:
                return
            if device_id not in self.connected_devices:
                self.connected_devices.add(device_id)
                logger.debug(
                    f'device {device_id} connected to shared Modbus client {self.identifier}'
                )
                await self.client.connect()
        else:
            await self.client.connect()

    async def disconnect(self, device_id: str):
        if self.shared:
            if device_id not in self.registered_devices:
                return
            if device_id in self.connected_devices:
                self.connected_devices.remove(device_id)
                logger.debug(
                    f'device {device_id} disconnected from shared Modbus client {self.identifier}'
                )
                if len(self.connected_devices) == 0:
                    logger.debug(
                        f'last device disconnected from shared Modbus client {self.identifier}'
                    )
                    await self.client.disconnect()
        else:
            await self.client.disconnect()

    def is_connected(self, device_id: str) -> bool:
        if self.shared:
            return (device_id in self.registered_devices) and (
                device_id is self.connected_devices
            )
        else:
            return self.client.is_connected()


# singleton objects
_global_shared_lock = Lock()
_global_shared_rtu_clients: dict[str, ModbusClientWrapper] = dict()


def register_shared_client(
    serial_port: str, parity: str, baudrate: int, device_id: str
) -> ModbusClientWrapper:
    global _global_shared_lock
    global _global_shared_rtu_clients
    with _global_shared_lock:
        client_wrapper = _global_shared_rtu_clients.get(serial_port)
        if client_wrapper is None:
            modbus_client = SGrModbusRTUClient(
                serial_port,
                parity,
                baudrate,
                BitOrder.BIG_ENDIAN,  # TODO bit order was missing, i just added want.
            )
            client_wrapper = ModbusClientWrapper(
                serial_port, modbus_client, shared=True
            )
            _global_shared_rtu_clients[serial_port] = client_wrapper
        client_wrapper.registered_devices.add(device_id)
        logger.debug(
            f'device {device_id} registered at shared Modbus client {client_wrapper.identifier}'
        )

        return client_wrapper


def unregister_shared_client(serial_port: str, device_id: str) -> None:
    global _global_shared_lock
    global _global_shared_rtu_clients
    with _global_shared_lock:
        client_wrapper = _global_shared_rtu_clients.get(serial_port)
        if client_wrapper is not None:
            client_wrapper.connected_devices.remove(device_id)
            client_wrapper.registered_devices.remove(device_id)
            if len(client_wrapper.registered_devices) == 0:
                try:
                    asyncio.get_event_loop().run_until_complete(
                        client_wrapper.client.disconnect()
                    )
                except Exception:
                    logger.warning(
                        f'could not disconnect shared transport {client_wrapper.identifier}'
                    )
                _global_shared_rtu_clients.pop(serial_port)
            logger.debug(
                f'device {device_id} unregistered from shared Modbus client {client_wrapper.identifier}'
            )



from asyncio import run
from asyncio.protocols import Protocol

from sgr_specification.v0.generic import DataDirectionProduct

from sgr_commhandler.api.data_point_api import DataPoint
from sgr_commhandler.api.data_types import DataTypes


class FunctionalProfile(Protocol):
    """
    Implements a functional profile.
    """

    def name(self) -> str:
        """
        Gets the functional profile name.

        Returns
        -------
        str
            the functional profile name
        """
        ...

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        """
        Gets all data points.

        Returns
        -------
        dict[tuple[str, str], DataPoint]
            all data points
        """
        ...

    def get_data_point(self, dp_name: str) -> DataPoint:
        """
        Gets a data point.

        Parameters
        ----------
        dp_name : str
            the data point name

        Returns
        -------
        DataPoint
            a data point
        """
        return self.get_data_points()[(self.name(), dp_name)]

    async def get_value_async(self) -> dict[str, DataPoint]:
        """
        Gets all data point values of the functional profile asynchronously.

        Returns
        -------
        dict[str, DataPoint]
            all data point values by name
        """
        return {
            key[1]: await dp.get_value_async()
            for key, dp in self.get_data_points().items()
        }

    def get_value(self) -> dict[str, DataPoint]:
        """
        Gets all data point values of the functional profile synchronously.

        Returns
        -------
        dict[str, DataPoint]
            all data point values by name
        """
        return run(self.get_value_async())

    def describe(
        self,
    ) -> tuple[str, dict[str, tuple[DataDirectionProduct, DataTypes]]]:
        """
        Describes the functional profile.

        Returns
        -------
        tuple[str, dict[str, tuple[DataDirectionProduct, DataTypes]]]
            the functional profile information
        """
        infos = map(lambda dp: dp.describe(), self.get_data_points().values())
        return self.name(), {dp[0][1]: (dp[1], dp[2]) for dp in infos}



from asyncio import run
from collections.abc import Callable
from typing import Any, Generic, Protocol, TypeVar

from sgr_specification.v0.generic import DataDirectionProduct, Units

from sgr_commhandler.api.data_types import DataTypes

"""Defines a generic data type."""
T = TypeVar('T')


class DataPointValidator(Protocol):
    """
    Defines an interface for data point validators.
    """

    def validate(self, value: Any) -> bool:
        """
        Validates the compatibility of a value.

        Parameters
        ----------
        value : Any
            the value to validate

        Returns
        -------
        bool
            true if value is compatible, false otherwise
        """
        ...

    def data_type(self) -> DataTypes:
        """
        Gets the validator's data type.

        Returns
        -------
        DataTypes
            the data type enumeration
        """
        ...

    def options(self) -> list[Any]:
        """
        Gets the validator's options.

        Returns
        -------
        list[Any]
            the options
        """
        return []


class DataPointProtocol(Protocol):
    """
    Defines an interface for data point protocols.
    """

    async def set_val(self, value: Any):
        """
        Writes the data point value.

        Parameters
        ----------
        value : Any
            the data point value to write
        """
        ...

    async def get_val(self, skip_cache: bool = False) -> Any:
        """
        Reads the data point value.

        Parameters
        ----------
        skip_cache : bool
            does not use cache if true
        
        Returns
        -------
        Any
            the data point value
        """
        ...

    def name(self) -> tuple[str, str]:
        """
        Gets the functional profile and data point names.

        Returns
        -------
        tuple[str, str]
            the functional profile and data point names as tuple
        """
        ...

    def direction(self) -> DataDirectionProduct:
        """
        Gets the data direction of the data point.

        Returns
        -------
        DataDirectionProduct
            the data point direction
        """
        ...

    def unit(self) -> Units:
        """
        Gets the unit of measurement of the data point.

        Returns
        -------
        Units
            the unit
        """
        ...

    def can_subscribe(self) -> bool:
        """
        Defines if subscribe() is allowed.

        Returns
        -------
        bool
            True if allowed, False otherwise
        """
        return False

    def subscribe(self, fn: Callable[[Any], None]):
        """
        Subscribes to changes of the data point value.

        Parameters
        ----------
        fn : Callable[[Any], None]
            the callback method
        """
        raise Exception('Unsupported operation')

    def unsubscribe(self):
        """
        Unsubscribes from changes of the data point value.
        """
        raise Exception('Unsupported operation')


class DataPoint(Generic[T]):
    """
    Implements a data point of a generic data type.
    """

    def __init__(
        self, protocol: DataPointProtocol, validator: DataPointValidator
    ):
        """
        Constructs a data point.

        Parameters
        ----------
        protocol : DataPointProtocol
            the underlying protocol
        validator : DataPointValidator
            the data point's value validator
        """
        self._protocol = protocol
        self._validator = validator

    def name(self) -> tuple[str, str]:
        """
        Gets the data point name.

        Returns
        -------
        tuple[str, str]
            the functional profile and data point name
        """
        return self._protocol.name()

    async def get_value_async(self) -> T:
        """
        Gets the data point value asynchronously.

        Returns
        -------
        T
            the data point value
        
        Raises
        ------
        Exception
            when read value is not compatible with data type
        """
        value = await self._protocol.get_val()
        if self._validator.validate(value):
            return value
        raise Exception(
            f'invalid value read from device, {value}, validator: {self._validator.data_type()}'
        )

    def get_value(self) -> T:
        """
        Gets the data point value synchronously.

        Returns
        -------
        T
            the data point value
        
        Raises
        ------
        Exception
            when read value is not compatible with data type
        """
        return run(self.get_value_async())

    async def set_value_async(self, value: T):
        """
        Sets the data point value asynchronously.

        Parameters
        ----------
        value: T
            the data point value
        """
        if self._validator.validate(value):
            return await self._protocol.set_val(value)
        raise Exception('invalid data to write to device')

    def set_value(self, value: T):
        """
        Sets the data point value synchronously.

        Parameters
        ----------
        value: T
            the data point value
        """
        return run(self.set_value_async(value))

    def subscribe(self, fn: Callable[[Any], None]):
        """
        Subscribes to data point value changes.

        Parameters
        ----------
        fn : Callable[[Any], None]
            the handler method
        """
        self._protocol.subscribe(fn)

    def unsubscribe(self):
        """
        Unsubscribes from data point value changes.
        """
        self._protocol.unsubscribe()

    def direction(self) -> DataDirectionProduct:
        """
        Gets the data point direction.

        Returns
        -------
        DataDirectionProduct
            the data point direction
        """
        return self._protocol.direction()

    def data_type(self) -> DataTypes:
        """
        Gets the data point data type.

        Returns
        -------
        DataTypes
            the data point data type
        """
        return self._validator.data_type()
    
    def unit(self) -> Units:
        """
        Gets the unit of measurement of the data point.

        Returns
        -------
        Units
            the unit
        """
        return self._protocol.unit()

    def describe(
        self,
    ) -> tuple[tuple[str, str], DataDirectionProduct, DataTypes, Units]:
        """
        Describes the data point.

        Returns
        -------
        tuple[tuple[str, str], DataDirectionProduct, DataTypes, Units]
            the data point information
        """
        return self.name(), self.direction(), self.data_type(), self.unit()

    def options(self) -> list[Any]:
        """
        Describes the data point options.

        Returns
        -------
        list[Any]
            the data point options
        """
        return self._validator.options()



__all__ = ["MessagingDataPoint", "MessagingFunctionalProfile", "SGrMessagingInterface"]

from .messaging_interface_async import (
    MessagingDataPoint,
    MessagingFunctionalProfile,
    SGrMessagingInterface,
)



import logging
from typing import Any

from pymodbus.payload import BinaryPayloadBuilder, BinaryPayloadDecoder
from sgr_specification.v0.product.modbus_types import ModbusDataType


logger = logging.getLogger(__name__)


class PayloadDecoder(BinaryPayloadDecoder):
    """
    Implements a Modbus payload decoder.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def decode(self, modbus_type: ModbusDataType, byte_count: int):
        """
        Decodes a data value.

        Parameters
        ----------
        modbus_type : ModbusDataType
            the Modbus data type to decode
        byte_count : int
            the number of bytes
        """

        # TODO enum, date_time
        if modbus_type.int8:
            return self.decode_8bit_int()
        elif modbus_type.int8_u:
            return self.decode_8bit_uint()
        elif modbus_type.int16:
            return self.decode_16bit_int()
        elif modbus_type.int16_u:
            return self.decode_16bit_uint()
        elif modbus_type.int32:
            return self.decode_32bit_int()
        elif modbus_type.int32_u:
            return self.decode_32bit_uint()
        elif modbus_type.int64:
            return self.decode_64bit_int()
        elif modbus_type.int64_u:
            return self.decode_64bit_uint()
        elif modbus_type.float32:
            return self.decode_32bit_float()
        elif modbus_type.float64:
            return self.decode_64bit_float()
        elif modbus_type.boolean:
            return bool(self.decode_8bit_uint())
        elif modbus_type.string:
            return self.decode_string(byte_count)
        else:
            raise ValueError("No supported modbus data type")


class PayloadBuilder(BinaryPayloadBuilder):
    """
    Implements a Modbus payload encoder.
    """

    def __init__(self, *args, **kwarg):
        super().__init__(*args, **kwarg)

    def sgr_encode(
        self, value: Any, modbus_type: ModbusDataType
    ) -> "PayloadBuilder":
        """
        Encodes data type value.

        Parameters
        ----------
        value : Any
            the value to encode
        modbus_type : ModbusDataType
            The Modbus data type to encode
        """
        # TODO enum, date_time
        if modbus_type.int8:
            self.add_8bit_int(int(value))
        elif modbus_type.int8_u:
            self.add_8bit_uint(int(value))
        elif modbus_type.int16:
            self.add_16bit_int(int(value))
        elif modbus_type.int16_u:
            self.add_16bit_uint(int(value))
        elif modbus_type.int32:
            self.add_32bit_int(int(value))
        elif modbus_type.int32_u:
            self.add_32bit_uint(int(value))
        elif modbus_type.int64:
            self.add_64bit_int(int(value))
        elif modbus_type.int64_u:
            self.add_64bit_uint(int(value))
        elif modbus_type.float32:
            self.add_32bit_float(float(value))
        elif modbus_type.float64:
            self.add_64bit_float(float(value))
        elif modbus_type.boolean:
            self.add_8bit_uint(bool(value))
        elif modbus_type.string:
            self.add_string(str(value))
        else:
            logger.error(f'Unsupported modbus type "{modbus_type}"')
        return self



from asyncio import run
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, Protocol

from sgr_specification.v0.generic import DataDirectionProduct, DeviceCategory
from sgr_specification.v0.product.product import DeviceFrame

from sgr_commhandler.api import DataPoint
from sgr_commhandler.api.configuration_parameter import (
    ConfigurationParameter,
    build_configuration_parameters,
)
from sgr_commhandler.api.data_types import DataTypes
from sgr_commhandler.api.functional_profile_api import FunctionalProfile


@dataclass
class DeviceInformation:
    """
    Implements a device information container.

    Attributes
    ----------
    name : str
        Defines the device name
    manufacturer : str
        Defines the manufacturer's name
    software_revision : str
        Defines the device software version
    hardware_revision : str
        Defines the device hardware version
    device_category : DeviceCategory
        Defines the device category
    is_local : bool
        Defines if the device is controlled locally or via cloud
    """
    name: str
    manufacturer: str
    software_revision: str
    hardware_revision: str
    device_category: DeviceCategory
    is_local: bool


class SGrBaseInterface(Protocol):
    """
    Defines an abstract base class for all SGr device interfaces.

    Attributes
    ----------
    device_frame : DeviceFrame
        the device specification
    configuration_parameters : list[ConfigurationParameter]
        the configuration parameters with default values
    device_information : DeviceInformation
        the device information
    functional_profiles : Mapping[str, FunctionalProfile]
        the configured functional profiles
    """
    device_frame: DeviceFrame
    configuration_parameters: list[ConfigurationParameter]
    device_information: DeviceInformation
    functional_profiles: Mapping[str, FunctionalProfile]

    def _inititalize_device(
        self, frame: DeviceFrame
    ):
        self.device_frame = frame
        self.configuration_parameters = build_configuration_parameters(
            frame.configuration_list
        )
        self.device_information = DeviceInformation(
            name=frame.device_name if frame.device_name else '',
            manufacturer=frame.manufacturer_name
            if frame.manufacturer_name
            else '',
            software_revision=frame.device_information.software_revision
            if frame.device_information
            and frame.device_information.software_revision
            else '',
            hardware_revision=frame.device_information.hardware_revision
            if frame.device_information
            and frame.device_information.hardware_revision
            else '',
            device_category=frame.device_information.device_category
            if frame.device_information
            and frame.device_information.device_category
            else DeviceCategory.DEVICE_INFORMATION,
            is_local=frame.device_information.is_local_control
            if frame.device_information
            and frame.device_information.is_local_control
            else False,
        )

    def connect(self):
        """
        Connects the device synchronously.
        """
        run(self.connect_async())

    async def connect_async(self):
        """
        Connects the device asynchronously.
        """
        ...

    def disconnect(self):
        """
        Disconnects the device synchronously.
        """
        run(self.disconnect_async())

    async def disconnect_async(self):
        """
        Disconnects the device asynchronously.
        """
        ...

    def is_connected(self) -> bool:
        """
        Gets the connection state.

        Returns
        -------
        bool
            the connection state
        """
        ...

    def get_functional_profile(
        self, functional_profile_name: str
    ) -> FunctionalProfile:
        """
        Gets a functional profile.

        Parameters
        ----------
        functional_profile_name : str
            the functional profile name

        Returns
        -------
        FunctionalProfile
            a functional profile
        """
        return self.functional_profiles[functional_profile_name]

    def get_data_point(self, dp: tuple[str, str]) -> DataPoint:
        """
        Gets a data point.

        Parameters
        ----------
        dp : tuple[str, str]
            the functional profile and data point name

        Returns
        -------
        DataPoint
            a data point
        """
        return self.get_functional_profile(dp[0]).get_data_point(dp[1])

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        """
        Gets all data points.

        Returns
        -------
        dict[tuple[str, str], DataPoint]
            all data points
        """
        data_points = {}
        for fp in self.functional_profiles.values():
            data_points.update(fp.get_data_points())
        return data_points

    def get_values(self) -> dict[tuple[str, str], Any]:
        """
        Gets all data point values synchronously.

        Returns
        -------
        dict[tuple[str, str], Any]
            all data point values
        """
        return run(self.get_values_async())

    async def get_values_async(self) -> dict[tuple[str, str], Any]:
        """
        Gets all data point values asynchronously.

        Returns
        -------
        dict[tuple[str, str], Any]
            all data point values
        """
        data = {}
        for fp in self.functional_profiles.values():
            data.update(
                {
                    (fp.name(), key): value
                    for key, value in (await fp.get_value_async()).items()
                }
            )
        return data

    def describe(
        self,
    ) -> tuple[
        str, dict[str, dict[str, tuple[DataDirectionProduct, DataTypes]]]
    ]:
        """
        Gets the device description and all data.

        Returns
        -------
        tuple[str, dict[str, dict[str, tuple[DataDirectionProduct, DataTypes]]]]
            a tuple of device name and all data point values
        """
        data = {}
        for fp in self.functional_profiles.values():
            key, dps = fp.describe()
            data[key] = dps
        return self.device_information.name, data



from typing import Optional

from sgr_specification.v0.generic import DataTypeProduct

from sgr_commhandler.api import DataPointValidator
from sgr_commhandler.validators.validator import (
    BitmapValidator,
    BooleanValidator,
    DateTimeValidator,
    EnumValidator,
    FloatValidator,
    IntValidator,
    StringValidator,
    JsonValidator
)


def build_validator(type: Optional[DataTypeProduct]) -> DataPointValidator:
    """
    Builds a data point validator from a data type specification.

    Parameters
    ----------
    type : Optional[DataTypeProduct]
        the data type specification

    Returns
    -------
    DataPointValidator
        a data point validator

    Raises
    ------
    Exception
        when data type has no supported validator
    """
    if type is None:
        raise Exception("Missing datatype")
    if type.int8:
        return IntValidator(8, signed=True)
    elif type.int16:
        return IntValidator(16, signed=True)
    elif type.int32:
        return IntValidator(32, signed=True)
    elif type.int64:
        return IntValidator(64, signed=True)
    elif type.enum:
        return EnumValidator(type.enum)
    elif type.int8_u:
        return IntValidator(8, signed=False)
    elif type.int16_u:
        return IntValidator(16, signed=False)
    elif type.int32_u:
        return IntValidator(32, signed=False)
    elif type.int64_u:
        return IntValidator(64, signed=False)
    elif type.float32:
        return FloatValidator(32)
    elif type.float64:
        return FloatValidator(64)
    elif type.string:
        return StringValidator()
    elif type.boolean:
        return BooleanValidator()
    elif type.bitmap:
        return BitmapValidator()
    elif type.date_time:
        return DateTimeValidator()
    elif type.json:
        return JsonValidator()
    raise Exception("unsupported validator")



from typing import Optional

from sgr_specification.v0.product import (
    ConfigurationList,
    ConfigurationListElement,
)


def build_configuration_parameters(params: Optional[ConfigurationList]) -> list['ConfigurationParameter']:
    """
    Constructs a configuration parameter list.

    Parameters
    ----------
    params : Optional[ConfigurationList]
        The configuration list of an EID
    
    Returns
    -------
    list[ConfigurationParameter]
        a list of configuration parameters
    """
    if params is None:
        return []
    return [
        ConfigurationParameter(x) for x in params.configuration_list_element
    ]


class ConfigurationParameter:
    """
    Implements an EID configuration parameter.
    """

    def __init__(self, parameter: ConfigurationListElement):
        """
        Constructs a configuration parameter.

        Parameters
        ----------
        parameter : ConfigurationListElement
            A configuration list element of the SGr specification
        """  
        translation = parameter.configuration_description
        self.label = translation[0].label
        self.name = parameter.name
        self.type = parameter.data_type
        self.description = translation[0].text_element
        self.default_value = parameter.default_value

    def __str__(self) -> str:
        """
        Converts to string.

        Returns
        ----------
        str
            A string
        """
        return f'{self.label} - {self.name}'

    def __repr__(self):
        """
        Gets an object description as string.

        Returns
        ----------
        str
            A string
        """
        return f'{self.label} - {self.name}'



__all__ = ["GenericDataPoint", "GenericFunctionalProfile", "SGrGenericInterface"]

from .generic_interface_async import (
    GenericDataPoint,
    GenericFunctionalProfile,
    SGrGenericInterface,
)



__all__ = ["ContactDataPoint", "ContactFunctionalProfile", "SGrContactInterface"]

from .contact_interface_async import (
    ContactDataPoint,
    ContactFunctionalProfile,
    SGrContactInterface,
)



__all__ = [
    "SGrBaseInterface",
    "FunctionalProfile",
    "DataPoint",
    "DataPointProtocol",
    "DataPointValidator",
    "DeviceInformation",
    "ConfigurationParameter",
]

from sgr_commhandler.api.configuration_parameter import ConfigurationParameter
from sgr_commhandler.api.data_point_api import (
    DataPoint,
    DataPointProtocol,
    DataPointValidator,
)
from sgr_commhandler.api.device_api import DeviceInformation, SGrBaseInterface
from sgr_commhandler.api.functional_profile_api import FunctionalProfile



import logging
import configparser
import re
import xmlschema
import importlib.resources
import sgr_schema
from collections.abc import Callable
from enum import Enum
from typing import cast

from sgr_specification.v0.product import DeviceFrame
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers import XmlParser

from sgr_commhandler.api.configuration_parameter import ConfigurationParameter, build_configuration_parameters
from sgr_commhandler.api.device_api import SGrBaseInterface
from sgr_commhandler.driver.contact.contact_interface_async import (
    SGrContactInterface,
)
from sgr_commhandler.driver.generic.generic_interface_async import (
    SGrGenericInterface,
)
from sgr_commhandler.driver.messaging.messaging_interface_async import (
    SGrMessagingInterface,
)
from sgr_commhandler.driver.modbus.modbus_interface_async import (
    SGrModbusInterface,
)
from sgr_commhandler.driver.rest.restapi_interface_async import SGrRestInterface

logger = logging.getLogger(__name__)


class SGrXmlSource(Enum):
    """Defines an EID XML source."""
    UNKNOWN = 1
    STRING = 2
    FILE = 3


class SGrPropertiesSource(Enum):
    """Defines an EID properties source."""
    UNKNOWN = 1
    DICT = 2
    FILE = 3


class SGrDeviceProtocol(Enum):
    """Defines a communication interface type."""
    MODBUS = 0
    RESTAPI = 1
    MESSAGING = 2
    CONTACT = 3
    GENERIC = 4
    UNKNOWN = 5


SGrInterfaces = (
    SGrRestInterface
    | SGrModbusInterface
    | SGrMessagingInterface
    | SGrContactInterface
    | SGrGenericInterface
)

device_builders: dict[
    SGrDeviceProtocol,
    Callable[
        [DeviceFrame],
        SGrInterfaces,
    ],
] = {
    SGrDeviceProtocol.MODBUS: lambda frame: SGrModbusInterface(
        frame, sharedRTU=True
    ),
    SGrDeviceProtocol.RESTAPI: lambda frame: SGrRestInterface(
        frame
    ),
    SGrDeviceProtocol.MESSAGING: lambda frame: SGrMessagingInterface(
        frame
    ),
    SGrDeviceProtocol.CONTACT: lambda frame: SGrContactInterface(
        frame
    ),
    SGrDeviceProtocol.GENERIC: lambda frame: SGrGenericInterface(
        frame
    ),
}


class DeviceBuilder:
    """
    Implements an SGr device builder with a fluent interface.
    """

    def __init__(self):
        """
        Constructs a new device builder.
        """
        self._eid_source: str | None = None
        self._properties_source: str | dict | None = None
        self._eid_type: SGrXmlSource = SGrXmlSource.UNKNOWN
        self._properties_type: SGrPropertiesSource = SGrPropertiesSource.UNKNOWN

    def build(self) -> SGrBaseInterface:
        """
        Constructs an SGr device instance from the builder.

        Returns
        -------
        SGrBaseInterface
            an instance of an SGr device
        """

        eid_content = self._load_eid_content()
        properties = self._load_properties()
        # parse EID - get configuration list
        frame = parse_device_frame(eid_content)
        # build final properties
        config_params = build_configuration_parameters(frame.configuration_list)
        properties = build_properties(config_params, properties)
        # parse EID - final
        eid_content = replace_variables(eid_content, properties)
        frame = parse_device_frame(eid_content)
        protocol = self._resolve_protocol(frame)
        return device_builders[protocol](frame)

    def _resolve_protocol(self, frame: DeviceFrame) -> SGrDeviceProtocol:
        if frame.interface_list is None:
            raise Exception('no device interface')
        if frame.interface_list.rest_api_interface:
            return SGrDeviceProtocol.RESTAPI
        elif frame.interface_list.modbus_interface:
            return SGrDeviceProtocol.MODBUS
        elif frame.interface_list.messaging_interface:
            return SGrDeviceProtocol.MESSAGING
        elif frame.interface_list.contact_interface:
            return SGrDeviceProtocol.CONTACT
        elif frame.interface_list.generic_interface:
            return SGrDeviceProtocol.GENERIC
        raise Exception('unsupported device interface')

    def eid_path(self, file_path: str) -> 'DeviceBuilder':
        """
        Sets the EID source to a file path.

        Parameters
        ----------
        file_path: str
            the path to the EID XML file
        
        Returns
        -------
        DeviceBuilder
            the same builder instance
        """
        self._eid_source = file_path
        self._eid_type = SGrXmlSource.FILE
        return self

    def eid(self, xml: str):
        """
        Sets the EID source to an XML string.

        Parameters
        ----------
        xml: str
            the EID XML content
        
        Returns
        -------
        DeviceBuilder
            the same builder instance
        """
        self._eid_source = xml
        self._eid_type = SGrXmlSource.STRING
        return self

    def properties_path(self, file_path: str):
        """
        Sets the EID properties to a file path.

        Parameters
        ----------
        file_path: str
            the path to the property file
        
        Returns
        -------
        DeviceBuilder
            the same builder instance
        """
        self._properties_type = SGrPropertiesSource.FILE
        self._properties_source = file_path
        return self

    def properties(self, properties: dict):
        """
        Sets the EID properties to a key-value map.

        Parameters
        ----------
        properties: dict
            the properties
        
        Returns
        -------
        DeviceBuilder
            the same builder instance
        """
        self._properties_type = SGrPropertiesSource.DICT
        self._properties_source = properties
        return self

    def _load_eid_content(self) -> str:
        if self._eid_source is None:
            raise Exception('No EID defined')
        content = ''
        if self._eid_type == SGrXmlSource.FILE:
            try:
                input_file = open(self._eid_source)
                content = input_file.read()
            except Exception:
                raise Exception('Invalid EID file path')
        elif self._eid_type == SGrXmlSource.STRING:
            content = self._eid_source
        return content

    def _load_properties(self) -> dict:
        if self._properties_source is None:
            logger.debug('no properties given')
            return {}
        if self._properties_type == SGrPropertiesSource.FILE:
            try:
                prop_path = str(self._properties_source)
                logger.debug(f'getting properties from file {prop_path}')
                config = configparser.RawConfigParser()
                config.optionxform = lambda option: option
                config.read(prop_path)
                properties = {}
                for section_name, section in config.items():
                    for param_name in section:
                        param_value = config.get(section_name, param_name)
                        properties[param_name] = param_value
                        logger.debug(f'found in property file: {param_name} = {param_value}')
                return properties
            except Exception as e:
                raise Exception(f'Error reading properties file {self._properties_source}: {e}')
        elif self._properties_type == SGrPropertiesSource.DICT:
            logger.debug('getting properties from dict')
            return cast(dict, self._properties_source)
        return {}


def parse_device_frame(content: str) -> DeviceFrame:
    """
    Parses EID XML content into a device frame.

    Parameters
    ----------
    content: str
        the EID XML content
    
    Returns
    -------
    DeviceFrame
        a device frame of the SGr specification
    """
    validate_schema(content)
    parser = XmlParser(context=XmlContext())
    return parser.from_string(content, DeviceFrame)


def replace_variables(content: str, parameters: dict) -> str:
    """
    Replaces parameter placeholders in EID XML content.

    Parameters
    ----------
    content: str
        the EID XML content
    parameters: dict
        the configuration parameters
    
    Returns
    -------
    str
        the updated EID XML content
    """
    for name, value in parameters.items():
        pattern = re.compile(r'{{' + str(name) + r'}}')
        content = pattern.sub(str(value), content)
        logger.debug(f'replaced parameter: {str(name)} = {str(value)}')
    return content


def build_properties(config: list[ConfigurationParameter], properties: dict) -> dict:
    """
    Builds EID configuration properties.
    Only parameters defined in the EID configuration list are kept.
    Parameters not defined in the properties are set to the default value.

    Parameters
    ----------
    config: list[ConfigurationParameter]
        the EID configuration parameters
    properties: dict
        the properties to configure
    
    Returns
    -------
    dict
        the final properties
    """
    final_properties = {}
    for config_param in config:
        prop_value = properties.get(config_param.name)
        if prop_value is not None:
            final_properties[config_param.name] = prop_value
        elif config_param.default_value is not None:
            final_properties[config_param.name] = config_param.default_value
        logger.debug(f'EID parameter: {config_param.name} = {final_properties[config_param.name]}')
    return final_properties


def validate_schema(content: str):
    """
    Validates EID XML content against the specification XML schema.

    Parameters
    ----------
    content: str
        the EID XML content
    """
    xsd_path = importlib.resources.files(sgr_schema).joinpath('SGrIncluder.xsd')
    xsd = xmlschema.XMLSchema(xsd_path)
    xsd.validate(content)



from setuptools import setup, find_packages


def read_requirements() -> list[str]:
    with open("./requirements.txt", "r") as file:
        return file.read().splitlines()


setup(
    name="SGrPythontks4r",  # use SGrPythontks4r until organization has been created, then change to sgr-commhandler
    version="0.3.1",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=read_requirements(),
    url="https://github.com/SmartGridready/SGrPython#readme",
    project_urls={
        "Issue Tracker": "https://github.com/SmartGridready/SGrPython/issues",
        "Source": "https://github.com/SmartGridready/SGrPython",
    },
    license="BSD-3",
    license_files=["LICENSE.md"],
    author="Robin Schoch",
    author_email="",
    description="SGr Commhandler",
    long_description="SGr Commhandler Library",
    long_description_content_type="text/plain",
    python_requires=">=3.9",
)



import json
import logging
import ssl
from io import UnsupportedOperation
from typing import Any, Callable, Optional
from urllib.parse import urlencode

import aiohttp
import certifi
import jmespath
from aiohttp import ClientConnectionError, ClientResponseError
from cachetools import TTLCache
from sgr_specification.v0.generic import DataDirectionProduct, Units
from sgr_specification.v0.generic.base_types import ResponseQueryType
from sgr_specification.v0.product import (
    DeviceFrame,
    HeaderEntry,
    HeaderList,
    HttpMethod,
)
from sgr_specification.v0.product import (
    RestApiDataPoint as RestApiDataPointSpec,
)
from sgr_specification.v0.product import (
    RestApiFunctionalProfile as RestApiFunctionalProfileSpec,
)
from sgr_specification.v0.product.rest_api_types import (
    ParameterList,
    RestApiServiceCall,
)

from sgr_commhandler.api import (
    DataPoint,
    DataPointProtocol,
    FunctionalProfile,
    SGrBaseInterface,
)
from sgr_commhandler.driver.rest.authentication import setup_authentication
from sgr_commhandler.validators import build_validator

logger = logging.getLogger(__name__)


def build_rest_data_point(
    data_point: RestApiDataPointSpec,
    functional_profile: RestApiFunctionalProfileSpec,
    interface: 'SGrRestInterface',
) -> DataPoint:
    protocol = RestDataPoint(data_point, functional_profile, interface)
    data_type = None
    if data_point.data_point and data_point.data_point.data_type:
        data_type = data_point.data_point.data_type
    validator = build_validator(data_type)
    return DataPoint(protocol, validator)


class RestResponse:
    def __init__(self, headers: HeaderList, body: Optional[str] = None):
        self.headers = headers
        self.body = body


class RestRequest:
    def __init__(
        self,
        method: HttpMethod,
        url: str,
        headers: HeaderList,
        query_parameters: ParameterList = ParameterList(),
        form_parameters: ParameterList = ParameterList(),
        body: Optional[str] = None,
    ):
        self.method = method
        self.url = url
        self.headers = headers
        self.query_parameters = query_parameters
        self.form_parameters = form_parameters
        self.body = body


class RestDataPoint(DataPointProtocol):
    """
    Implements a data point of a REST API interface.
    """

    def __init__(
        self,
        dp_spec: RestApiDataPointSpec,
        fp_spec: RestApiFunctionalProfileSpec,
        interface: 'SGrRestInterface',
    ):
        self._dp_spec = dp_spec
        self._fp_spec = fp_spec

        dp_config = self._dp_spec.rest_api_data_point_configuration
        if not dp_config:
            raise Exception('REST service call configuration missing')

        self._read_call: RestApiServiceCall = RestApiServiceCall()
        self._write_call: RestApiServiceCall = RestApiServiceCall()

        if len(dp_config.rest_api_read_service_call) > 0:
            service_call = dp_config.rest_api_read_service_call[0]
            self._read_call = RestApiServiceCall(
                request_method=(
                    service_call.request_method
                    if service_call.request_method
                    else HttpMethod.GET
                ),
                request_path=(
                    service_call.request_path
                    if service_call.request_path
                    else ''
                ),
                request_header=(
                    service_call.request_header
                    if service_call.request_header
                    else HeaderList()
                ),
                request_query=(
                    service_call.request_query
                    if service_call.request_query
                    else ParameterList()
                ),
                request_form=(
                    service_call.request_form
                    if service_call.request_form
                    else ParameterList()
                ),
                response_query=(
                    service_call.response_query
                    if service_call.response_query
                    else None
                ),
            )
        elif dp_config.rest_api_service_call:
            # old spec, for compatibility reasons
            self._read_call = RestApiServiceCall(
                request_method=(
                    dp_config.rest_api_service_call.request_method
                    if dp_config.rest_api_service_call.request_method
                    else HttpMethod.GET
                ),
                request_path=(
                    dp_config.rest_api_service_call.request_path
                    if dp_config.rest_api_service_call.request_path
                    else ''
                ),
                request_header=(
                    dp_config.rest_api_service_call.request_header
                    if dp_config.rest_api_service_call.request_header
                    else HeaderList()
                ),
                request_query=(
                    dp_config.rest_api_service_call.request_query
                    if dp_config.rest_api_service_call.request_query
                    else ParameterList()
                ),
                request_form=(
                    dp_config.rest_api_service_call.request_form
                    if dp_config.rest_api_service_call.request_form
                    else ParameterList()
                ),
                response_query=(
                    dp_config.rest_api_service_call.response_query
                    if dp_config.rest_api_service_call.response_query
                    else None
                ),
            )

        if len(dp_config.rest_api_write_service_call) > 0:
            service_call = dp_config.rest_api_write_service_call[0]
            self._write_call = RestApiServiceCall(
                request_method=(
                    service_call.request_method
                    if service_call.request_method
                    else HttpMethod.GET
                ),
                request_path=(
                    service_call.request_path
                    if service_call.request_path
                    else ''
                ),
                request_header=(
                    service_call.request_header
                    if service_call.request_header
                    else HeaderList()
                ),
                request_query=(
                    service_call.request_query
                    if service_call.request_query
                    else ParameterList()
                ),
                request_form=(
                    service_call.request_form
                    if service_call.request_form
                    else ParameterList()
                ),
                response_query=(
                    service_call.response_query
                    if service_call.response_query
                    else None
                ),
            )

        if not self._read_call and not self._write_call:
            raise Exception('No REST service call configured')

        self._fp_name = ''
        if (
            fp_spec.functional_profile is not None
            and fp_spec.functional_profile.functional_profile_name is not None
        ):
            self._fp_name = fp_spec.functional_profile.functional_profile_name

        self._dp_name = ''
        if (
            dp_spec.data_point is not None
            and dp_spec.data_point.data_point_name is not None
        ):
            self._dp_name = dp_spec.data_point.data_point_name

        self._interface = interface

    def name(self) -> tuple[str, str]:
        return self._fp_name, self._dp_name

    async def get_val(self, skip_cache: bool = False):
        if not self._read_call:
            raise Exception('No read call')

        # TODO das scheint auch noch falsch muss ein reqeust jeweils alle parameter haben?
        request = RestRequest(
            self._read_call.request_method,
            f'{self._interface.base_url}{self._read_call.request_path}',
            self._read_call.request_header,
            self._read_call.request_query,
            self._read_call.request_form,
            self._read_call.request_body,
        )
        response = await self._interface.execute_request(request, skip_cache)
        if not response.body:
            return None
        if (
            self._read_call.response_query
            and self._read_call.response_query.query_type
            == ResponseQueryType.JMESPATH_EXPRESSION
        ):
            query_expression = self._read_call.response_query.query
            return jmespath.search(query_expression, json.loads(response.body))
        ret_value = response.body

        # convert to DP units
        if (
            self._dp_spec.data_point
            and self._dp_spec.data_point.unit_conversion_multiplicator
            and self._dp_spec.data_point.unit_conversion_multiplicator != 1.0
        ):
            ret_value = (
                float(ret_value)
                * self._dp_spec.data_point.unit_conversion_multiplicator
            )

        return ret_value

    async def set_val(self, value: Any):
        if not self._write_call:
            raise Exception('No write call')

        # convert to device units
        if (
            self._dp_spec.data_point is None
            or self._dp_spec.data_point.unit_conversion_multiplicator
            and self._dp_spec.data_point.unit_conversion_multiplicator != 1.0
        ):
            value = (
                float(value)
                / self._dp_spec.data_point.unit_conversion_multiplicator
            )

        # TODO auch hier scheint alles no ein bisschen fehlerhaft
        # replace [[value]] placeholder
        request = RestRequest(
            self._write_call.request_method,
            f'{self._interface.base_url}{self._write_call.request_path}',
            self._write_call.request_header,
            self._write_call.request_query,
            self._write_call.request_form,
            body=str(self._write_call.request_body).replace(
                '[[value]]', str(value)
            )
            if self._write_call.request_body
            else None,
        )
        # TODO use response body
        await self._interface.execute_request(request, skip_cache=True)

    def direction(self) -> DataDirectionProduct:
        if (
            self._dp_spec.data_point is None
            or self._dp_spec.data_point.data_direction is None
        ):
            raise Exception('missing data direction')
        return self._dp_spec.data_point.data_direction
    
    def unit(self) -> Units:
        if (
            self._dp_spec.data_point is None
            or self._dp_spec.data_point.unit is None
        ):
            return Units.NONE
        return self._dp_spec.data_point.unit

    def subscribe(self, fn: Callable[[Any], None]):
        raise UnsupportedOperation(
            'subscribe is no available for rest data point'
        )

    def unsubscribe(self):
        raise UnsupportedOperation(
            'unsubscribe is no available for rest data point'
        )


class RestFunctionalProfile(FunctionalProfile):
    """
    Implements a functional profile of a REST API interface.
    """

    def __init__(
        self,
        fp_spec: RestApiFunctionalProfileSpec,
        interface: 'SGrRestInterface',
    ):
        self._fp_spec = fp_spec
        self._interface = interface

        raw_dps = []
        if (
            self._fp_spec.data_point_list
            and self._fp_spec.data_point_list.data_point_list_element
        ):
            raw_dps = self._fp_spec.data_point_list.data_point_list_element

        dps = [
            build_rest_data_point(dp, self._fp_spec, self._interface)
            for dp in raw_dps
        ]

        self._data_points = {dp.name(): dp for dp in dps}

    def name(self) -> str:
        if (
            self._fp_spec.functional_profile
            and self._fp_spec.functional_profile.functional_profile_name
        ):
            return self._fp_spec.functional_profile.functional_profile_name
        return ''

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        return self._data_points


class SGrRestInterface(SGrBaseInterface):
    """
    Implements a REST API device interface.
    """

    def __init__(
        self, frame: DeviceFrame
    ):
        self._inititalize_device(frame)
        self._session = None
        self._ssl_context = ssl.create_default_context(cafile=certifi.where())
        self._connector = aiohttp.TCPConnector(ssl=self._ssl_context)
        self._cache = TTLCache(maxsize=100, ttl=5)

        if (
            self.device_frame.interface_list
            and self.device_frame.interface_list
            and self.device_frame.interface_list.rest_api_interface
        ):
            self._raw_interface = self.device_frame.interface_list.rest_api_interface
        else:
            raise Exception('No REST interface')
        desc = self._raw_interface.rest_api_interface_description
        if desc is None:
            raise Exception('No REST interface description')
        self.base_url = desc.rest_api_uri
        if self.base_url is None:
            raise Exception('Invalid base URL')

        raw_fps = []
        if (
            self._raw_interface.functional_profile_list
            and self._raw_interface.functional_profile_list.functional_profile_list_element
        ):
            raw_fps = self._raw_interface.functional_profile_list.functional_profile_list_element
        fps = [RestFunctionalProfile(profile, self) for profile in raw_fps]
        self.functional_profiles = {fp.name(): fp for fp in fps}

    def is_connected(self):
        return self._session is not None and not self._session.closed

    async def disconnect_async(self):
        if self._session is not None and not self._session.closed:
            await self._session.close()
        self._session = None

    async def connect_async(self):
        if self._session is None or self._session.closed:
            self._session = aiohttp.ClientSession(connector=self._connector)
            await self.authenticate()

    async def authenticate(self):
        if self._session:
            await setup_authentication(self._raw_interface, self._session)

    async def execute_request(
        self, request: RestRequest, skip_cache: bool
    ) -> RestResponse:
        try:
            if self._session is None:
                raise Exception('no connection to device established')
            # All headers into dictionary
            request_headers = {
                header_entry.header_name: header_entry.value
                for header_entry in request.headers.header
            }

            # All query parameters into dictionary
            query_parameters = {
                param_entry.name: param_entry.value
                for param_entry in request.query_parameters.parameter
            }

            request_body: Optional[str] = request.body

            # All form parameters into dictionary
            form_parameters = {
                param_entry.name: param_entry.value
                for param_entry in request.form_parameters.parameter
            }
            # override body
            if len(form_parameters) > 0:
                request_body = urlencode(form_parameters)
                request_headers['Content-Type'] = (
                    'application/x-www-form-urlencoded'
                )

            cache_key = (frozenset(request_headers), request.url)
            if not skip_cache and cache_key in self._cache:
                return self._cache[cache_key]

            async with self._session.request(
                request.method.value,
                request.url,
                headers=request_headers,
                params=query_parameters,
                data=request_body,
            ) as req:
                req.raise_for_status()  # Raises an HTTPError if the HTTP request returned an unsuccessful status code
                logger.info(f'execute_request status: {req.status}')
                res_body = await req.text()

                sgr_headers = []
                for name, value in req.headers.items():
                    sgr_headers.append(
                        HeaderEntry(header_name=name, value=value)
                    )
                header_list = HeaderList(header=sgr_headers)
                response = RestResponse(headers=header_list, body=res_body)
                if not skip_cache:
                    self._cache[cache_key] = response
                return response

        except ClientResponseError as e:
            logger.error(f'HTTP error occurred: {e}')
            raise e
        except ClientConnectionError as e:
            logger.error(f'Connection error occurred: {e}')
            raise e
        except json.JSONDecodeError as e:
            logger.error(f'Failed to decode JSON: {e}')
            raise e
        except Exception as e:
            logger.error(f'An unexpected error occurred: {e}')
            raise e



import logging
from collections.abc import Callable
from typing import Any

from sgr_specification.v0.generic import DataDirectionProduct, Units
from sgr_specification.v0.product import (
    DeviceFrame,
)
from sgr_specification.v0.product import (
    MessagingDataPoint as MessagingDataPointSpec,
)
from sgr_specification.v0.product import (
    MessagingFunctionalProfile as MessagingFunctionalProfileSpec,
)

from sgr_commhandler.api import (
    DataPoint,
    DataPointProtocol,
    FunctionalProfile,
    SGrBaseInterface,
)
from sgr_commhandler.validators import build_validator

logger = logging.getLogger(__name__)


def build_messaging_data_point(
    data_point: MessagingDataPointSpec,
    functional_profile: MessagingFunctionalProfileSpec,
    interface: 'SGrMessagingInterface',
) -> DataPoint:
    """
    Builds a data point of a messaging interface.
    """
    protocol = MessagingDataPoint(data_point, functional_profile, interface)
    data_type = None
    if data_point.data_point and data_point.data_point.data_type:
        data_type = data_point.data_point.data_type
    validator = build_validator(data_type)
    return DataPoint(protocol, validator)


class MessagingDataPoint(DataPointProtocol):
    """
    Implements a data point of a messaging interface.
    """

    def __init__(
        self,
        dp_spec: MessagingDataPointSpec,
        fp_spec: MessagingFunctionalProfileSpec,
        interface: 'SGrMessagingInterface',
    ):
        self._dp_spec = dp_spec
        self._fp_spec = fp_spec

        dp_config = self._dp_spec.messaging_data_point_configuration
        if not dp_config:
            raise Exception('Messaging data point configuration missing')

        self._fp_name = ''
        if (
            fp_spec.functional_profile is not None
            and fp_spec.functional_profile.functional_profile_name is not None
        ):
            self._fp_name = fp_spec.functional_profile.functional_profile_name

        self._dp_name = ''
        if (
            dp_spec.data_point is not None
            and dp_spec.data_point.data_point_name is not None
        ):
            self._dp_name = dp_spec.data_point.data_point_name

        self._interface = interface

    def name(self) -> tuple[str, str]:
        return self._fp_name, self._dp_name

    async def get_val(self, skip_cache: bool = False):
        raise Exception('Not implemented')

    async def set_val(self, value: Any):
        raise Exception('Not implemented')

    def direction(self) -> DataDirectionProduct:
        if (
            self._dp_spec.data_point is None
            or self._dp_spec.data_point.data_direction is None
        ):
            raise Exception('missing data direction')
        return self._dp_spec.data_point.data_direction

    def unit(self) -> Units:
        if (
            self._dp_spec.data_point is None
            or self._dp_spec.data_point.unit is None
        ):
            return Units.NONE
        return self._dp_spec.data_point.unit

    def can_subscribe(self) -> bool:
        return True

    def subscribe(self, fn: Callable[[Any], None]):
        # TODO implement
        raise Exception('not implemented yet')

    def unsubscribe(self):
        # TODO implement
        raise Exception('not implemented yet')


class MessagingFunctionalProfile(FunctionalProfile):
    """
    Implements a functional profile of a messaging interface.
    """

    def __init__(
        self,
        fp_spec: MessagingFunctionalProfileSpec,
        interface: 'SGrMessagingInterface',
    ):
        self._fp_spec = fp_spec
        self._interface = interface

        raw_dps = []
        if (
            self._fp_spec.data_point_list
            and self._fp_spec.data_point_list.data_point_list_element
        ):
            raw_dps = self._fp_spec.data_point_list.data_point_list_element

        dps = [
            build_messaging_data_point(dp, self._fp_spec, self._interface)
            for dp in raw_dps
        ]

        self._data_points = {dp.name(): dp for dp in dps}

    def name(self) -> str:
        if (
            self._fp_spec.functional_profile
            and self._fp_spec.functional_profile.functional_profile_name
        ):
            return self._fp_spec.functional_profile.functional_profile_name
        return ''

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        return self._data_points


class SGrMessagingInterface(SGrBaseInterface):
    """
    SmartGridready External Interface Class for Messaging Protocols
    """

    def __init__(
        self, frame: DeviceFrame
    ):
        self._inititalize_device(frame)

        if (
            self.device_frame.interface_list
            and self.device_frame.interface_list
            and self.device_frame.interface_list.messaging_interface
        ):
            self._raw_interface = (
                self.device_frame.interface_list.messaging_interface
            )
        else:
            raise Exception('No messaging interface')
        desc = self._raw_interface.messaging_interface_description
        if desc is None:
            raise Exception('No messaging interface description')

        # TODO configure interface

        raw_fps = []
        if (
            self._raw_interface.functional_profile_list
            and self._raw_interface.functional_profile_list.functional_profile_list_element
        ):
            raw_fps = self._raw_interface.functional_profile_list.functional_profile_list_element
        fps = [MessagingFunctionalProfile(profile, self) for profile in raw_fps]
        self.functional_profiles = {fp.name(): fp for fp in fps}

    def is_connected(self):
        return False

    async def disconnect_async(self):
        # TODO implement
        pass

    async def connect_async(self):
        # TODO implement
        pass



import logging
import threading
from abc import ABC
from typing import Any, Optional

from pymodbus import FramerType
from pymodbus.client import AsyncModbusSerialClient, AsyncModbusTcpClient
from pymodbus.client.base import ModbusBaseClient
from pymodbus.constants import Endian
from sgr_specification.v0.product.modbus_types import BitOrder, ModbusDataType

from sgr_commhandler.driver.modbus.payload_decoder import (
    PayloadBuilder,
    PayloadDecoder,
)

logger = logging.getLogger(__name__)


class SGrModbusClient(ABC):
    """
    Defines an abstract base class for Modbus clients.
    """

    def __init__(self, endianness: BitOrder, addr_offset: int):
        self._lock = threading.Lock()
        self._client: Optional[ModbusBaseClient] = None
        self._byte_order: Endian = (
            Endian.BIG
            if endianness is None or endianness == BitOrder.BIG_ENDIAN
            else Endian.LITTLE
        )
        self._word_order: Endian = (
            Endian.LITTLE
            if endianness
            in {BitOrder.CHANGE_WORD_ORDER, BitOrder.CHANGE_DWORD_ORDER}
            else Endian.BIG
        )
        self._addr_offset = addr_offset

    async def connect(self): ...

    async def disconnect(self): ...

    def is_connected(self) -> bool: ...

    async def write_holding_registers(
        self, slave_id: int, address: int, data_type: ModbusDataType, value: Any
    ) -> None:
        """
        Encodes value to be written to holding register address.

        Parameters
        ----------
        slave_id : int
            The slave ID of the device
        address : int
            The address to read from and decode
        data_type : ModbusDataType
            The modbus type to encode
        value : Any
            The value to be written
        """
        if self._client is None:
            raise Exception('Client not initialized')
        builder = PayloadBuilder(
            byteorder=self._byte_order, wordorder=self._word_order
        )
        builder.sgr_encode(value, data_type)
        with self._lock:
            await self._client.write_registers(
                address+self._addr_offset, builder.to_registers(), slave=slave_id
            )

    async def write_coils(
        self, slave_id: int, address: int, data_type: ModbusDataType, value: Any
    ) -> None:
        """
        Encodes value to be written to coil address.

        Parameters
        ----------
        slave_id : int
            The slave ID of the device
        address : int
            The address to read from and decode
        data_type : ModbusDataType
            The modbus type to encode
        value : Any
            The value to be written
        """
        if self._client is None:
            raise Exception('Client not initialized')
        builder = PayloadBuilder(
            byteorder=self._byte_order, wordorder=self._word_order
        )
        builder.sgr_encode(value, data_type)
        with self._lock:
            await self._client.write_coils(
                address+self._addr_offset, builder.to_coils(), slave=slave_id
            )

    async def read_input_registers(
        self, slave_id: int, address: int, size: int, data_type: ModbusDataType
    ) -> Any:
        """
        Reads input registers and decodes the value.

        Parameters
        ----------
        slave_id : int
            The slave ID of the device
        address : int
            The address to read from and decode
        size : int
            The number of registers to read
        data_type : ModbusDataType
            The modbus type to decode
        
        Returns
        -------
        Any
            Decoded value
        """
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            response = await self._client.read_input_registers(
                address+self._addr_offset, count=size, slave=slave_id
            )
        if response and not response.isError():
            decoder = PayloadDecoder.fromRegisters(
                response.registers,
                byteorder=self._byte_order,
                wordorder=self._word_order,
            )
            return decoder.decode(data_type, 0)

    async def read_holding_registers(
        self, slave_id: int, address: int, size: int, data_type: ModbusDataType
    ) -> Any:
        """
        Reads holding registers and decodes the value.

        Parameters
        ----------
        slave_id : int
            The slave ID of the device
        address : int
            The address to read from and decode
        size : int
            The number of registers to read
        data_type : ModbusDataType
            The modbus type to decode
        
        Returns
        -------
        Any
            Decoded value
        """
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            response = await self._client.read_holding_registers(
                address+self._addr_offset, count=size, slave=slave_id
            )
        if response and not response.isError():
            decoder = PayloadDecoder.fromRegisters(
                response.registers,
                byteorder=self._byte_order,
                wordorder=self._word_order,
            )
            return decoder.decode(data_type, 0)

    async def read_coils(
        self, slave_id: int, address: int, size: int, data_type: ModbusDataType
    ) -> Any:
        """
        Reads coils and decodes the value.

        Parameters
        ----------
        slave_id : int
            The slave ID of the device
        address : int
            The address to read from and decode
        size : int
            The number of coils to read
        data_type : ModbusDataType
            The modbus type to decode
        
        Returns
        -------
        Any
            Decoded value
        """
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            response = await self._client.read_coils(
                address+self._addr_offset, count=size, slave=slave_id
            )
        if response and not response.isError():
            decoder = PayloadDecoder.fromCoils(
                response.bits,
                byteorder=self._byte_order,
                _wordorder=self._word_order,
            )
            return decoder.decode(data_type, 0)

    async def read_discrete_inputs(
        self, slave_id: int, address: int, size: int, data_type: ModbusDataType
    ) -> Any:
        """
        Reads discrete inputs and decodes the value.

        Parameters
        ----------
        slave_id : int
            The slave ID of the device
        address : int
            The address to read from and decode
        size : int
            The number of inputs to read
        data_type : ModbusDataType
            The modbus type to decode
        
        Returns
        -------
        Any
            Decoded value
        """
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            response = await self._client.read_discrete_inputs(
                address+self._addr_offset, count=size, slave=slave_id
            )
        if response and not response.isError():
            decoder = PayloadDecoder.fromCoils(
                response.bits,
                byteorder=self._byte_order,
                _wordorder=self._word_order,
            )
            return decoder.decode(data_type, 0)

    # TODO Implement block transfers and remove this method
    async def _mult_value_decoder(
        self,
        addr: int,
        size: int,
        data_type: ModbusDataType,
        register_type: str,
        slave_id: int,
    ) -> Optional[tuple[float,float,float]]:
        """
        Reads register and decodes the value.

        Parameters
        ----------
        addr : int
            The address to read from and decode
        size : int
            The number of registers to read
        data_type : ModbusDataType
            The modbus type to decode
        register_type : str
            The register type
        slave_id : int
            The slave ID of the device 

        Returns
        -------
        Optional[tuple[float,float,float]]
            Decoded float tuple
        """
        if self._client is None:
            raise Exception('Client not initialized')
        if register_type == 'HoldRegister':
            reg = self._client.read_holding_registers(
                addr, count=size, slave=slave_id
            )
        else:
            reg = self._client.read_input_registers(
                addr, count=size, slave=slave_id
            )
        reg = await reg
        decoder = PayloadDecoder.fromRegisters(
            reg.registers,
            byteorder=self._byte_order,
            wordorder=self._word_order,
        )
        # logger.debug(decoder.decode('float32', 0))
        if not reg.isError():
            # logger.debug(decoder.decode(data_type, 0))
            indexes = [size // 3 * 0, size // 3 * 1, size // 3 * 2]
            l1 = decoder.decode(data_type, indexes[0])
            l2 = decoder.decode(data_type, indexes[1])
            l3 = decoder.decode(data_type, indexes[2])
            return l1, l2, l3


class SGrModbusTCPClient(SGrModbusClient):
    """
    Implements a Modbus TCP client.
    """

    def __init__(self, ip: str, port: int, endianness: BitOrder = BitOrder.BIG_ENDIAN, addr_offset: int = 0):
        super().__init__(endianness, addr_offset)
        """
        Creates client.

        Parameters
        ----------
        ip : str
            The host to connect to (default 127.0.0.1)
        port : int
            The modbus port to connect to (default 502)
        endianness : BitOrder
            The data endianness
        addr_offset : int
            The address offset
        """
        self._ip = ip
        self._port = port
        self._client = AsyncModbusTcpClient(
            host=ip,
            port=port,
            timeout=1,
            retries=0,
            reconnect_delay=5000,
            reconnect_delay_max=30000,
        )

    async def connect(self):
        if self._client is None:
            raise Exception('Client not initialized')

        with self._lock:
            await self._client.connect()
            logger.debug('Connected to ModbusTCP on ip: ' + self._ip)

    async def disconnect(self):
        if self._client is None:
            return
        with self._lock:
            self._client.close()
            logger.debug('Disconnected from ModbusTCP on ip: ' + self._ip)

    def is_connected(self) -> bool:
        return self._client is not None and self._client.connected


class SGrModbusRTUClient(SGrModbusClient):
    def __init__(
        self, serial_port: str, parity: str, baudrate: int, endianness: BitOrder = BitOrder.BIG_ENDIAN, addr_offset: int = 0
    ):
        super().__init__(endianness, addr_offset)
        """
        Creates client.

        Parameters
        ----------
        serial_port : str
            The serial port to connect to (e.g. COM1)
        parity : str
            The serial parity (e.g. EVEN)
        baudrate : int
            The serial baudrate (e.g. 19200)
        endianness : BitOrder
            The data endianness
        addr_offset : int
            The address offset
        """
        self._serial_port = serial_port
        self._client = AsyncModbusSerialClient(
            port=serial_port,
            framer=FramerType.RTU,
            parity=parity,
            baudrate=baudrate,
        )  # changed source: https://stackoverflow.com/questions/58773476/why-do-i-get-pymodbus-modbusioexception-on-20-of-attempts

    async def connect(self):
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            await self._client.connect()
            logger.debug(
                'Connected to ModbusRTU on serial port: ' + self._serial_port
            )

    async def disconnect(self):
        if self._client is None:
            raise Exception('Client not initialized')
        with self._lock:
            self._client.close()
            logger.debug(
                'Disconnected from ModbusRTU on serial port: '
                + self._serial_port
            )

    def is_connected(self) -> bool:
        return self._client is not None and self._client.connected



import logging
import random
import string
from typing import Any, Optional

from sgr_specification.v0.generic import DataDirectionProduct, Parity
from sgr_specification.v0.generic.base_types import DataTypeProduct, Units
from sgr_specification.v0.product import (
    DeviceFrame,
)
from sgr_specification.v0.product import (
    ModbusDataPoint as ModbusDataPointSpec,
)
from sgr_specification.v0.product import (
    ModbusFunctionalProfile as ModbusFunctionalProfileSpec,
)
from sgr_specification.v0.product.modbus_types import (
    BitOrder,
    ModbusDataType,
    ModbusInterfaceDescription,
    ModbusInterfaceSelection,
    ModbusRtu,
    ModbusTcp,
    RegisterType,
)

from sgr_commhandler.api import (
    DataPoint,
    DataPointProtocol,
    FunctionalProfile,
    SGrBaseInterface,
)
from sgr_commhandler.driver.modbus.modbus_client_async import (
    SGrModbusRTUClient,
    SGrModbusTCPClient,
)
from sgr_commhandler.driver.modbus.shared_client import (
    ModbusClientWrapper,
    register_shared_client,
    unregister_shared_client,
)
from sgr_commhandler.utils import value_util
from sgr_commhandler.validators import build_validator

logger = logging.getLogger(__name__)


def get_rtu_slave_id(modbus_rtu: ModbusRtu) -> int:
    """
    returns the selected slave address
    """
    if modbus_rtu.slave_addr is None:
        raise Exception('No RTU slave address configured')
    return int(modbus_rtu.slave_addr)


def get_tcp_slave_id(modbus_tcp: ModbusTcp) -> int:
    """
    returns the selected slave address
    """
    if modbus_tcp.slave_id is None:
        raise Exception('No slave id configured')
    return int(modbus_tcp.slave_id)


def get_endian(modbus: ModbusInterfaceDescription) -> BitOrder:
    """
    returns the byte order.
    """
    if modbus.bit_order:
        return modbus.bit_order
    return BitOrder.BIG_ENDIAN


def get_address_offset(modbus: ModbusInterfaceDescription) -> int:
    """
    returns the address offset.
    is 0 by default and -1 when first register address is 1.
    """
    return -1 if modbus.first_register_address_is_one is not None and modbus.first_register_address_is_one else 0


def get_tcp_address(modbus_tcp: ModbusTcp) -> str:
    """
    returns the selected ip address.
    """
    if modbus_tcp.address is None:
        raise Exception('no modbus tcp address configured')
    return modbus_tcp.address


def get_tcp_port(modbus_tcp: ModbusTcp) -> int:
    """
    returns the selected ip port.
    """
    if modbus_tcp.port is None:
        raise Exception('no modbus tcp port configured')
    return int(modbus_tcp.port)


def get_rtu_serial_port(modbus_rtu: ModbusRtu) -> str:
    """
    returns the selected serial port.
    """
    if modbus_rtu.port_name is None:
        raise Exception('no modbus rtu port name configured')
    return modbus_rtu.port_name


def get_rtu_baudrate(modbus_rtu: ModbusRtu) -> int:
    """
    returns the selected baudrate.
    """
    if modbus_rtu.baud_rate_selected is None:
        raise Exception('no modbus rtu baud_rate_selected name configured')
    return int(modbus_rtu.baud_rate_selected)


def get_rtu_parity(modbus_rtu: ModbusRtu) -> str:
    """
    returns the parity.
    """
    parity = modbus_rtu.parity_selected
    match parity:
        case Parity.NONE:
            return 'N'
        case Parity.EVEN:
            return 'E'
        case Parity.ODD:
            return 'O'
        case _:
            raise NotImplementedError


def build_modbus_data_point(
    data_point: ModbusDataPointSpec,
    functional_profile: ModbusFunctionalProfileSpec,
    interface: 'SGrModbusInterface',
) -> DataPoint:
    """
    Builds a data point of a Modbus interface.
    """

    protocol = ModbusDataPoint(data_point, functional_profile, interface)
    validator = build_validator(
        data_point.data_point.data_type if data_point.data_point else None
    )
    return DataPoint(protocol, validator)


def is_integer_type(data_type: DataTypeProduct | ModbusDataType) -> bool:
    """
    Checks if a data type is an integer.

    Returns
    -------
    bool
        True if integer, False otherwise
    """

    return any(
        (
            data_type.int8 is not None,
            data_type.int8_u is not None,
            data_type.int16 is not None,
            data_type.int16_u is not None,
            data_type.int32 is not None,
            data_type.int32_u is not None,
            data_type.int64 is not None,
            data_type.int64_u is not None,
        )
    )


def is_float_type(data_type: DataTypeProduct | ModbusDataType) -> bool:
    """
    Checks if a data type is a floating point value.

    Returns
    -------
    bool
        True if integer, False otherwise
    """

    return data_type.float32 is not None or data_type.float64 is not None


class ModbusDataPoint(DataPointProtocol):
    """
    Implements a data point of a Modbus interface.
    """

    def __init__(
        self,
        dp_spec: ModbusDataPointSpec,
        fp_spec: ModbusFunctionalProfileSpec,
        interface: 'SGrModbusInterface',
    ):
        self._dp_spec = dp_spec
        self._fp_spec = fp_spec
        self._interface = interface

        self._dp_name: str = ''
        if (
            self._dp_spec.data_point
            and self._dp_spec.data_point.data_point_name
        ):
            self._dp_name = self._dp_spec.data_point.data_point_name

        self._fp_name: str = ''
        if (
            self._fp_spec.functional_profile
            and self._fp_spec.functional_profile.functional_profile_name
        ):
            self._fp_name = (
                self._fp_spec.functional_profile.functional_profile_name
            )

        self._address: int = -1
        if (
            self._dp_spec.modbus_data_point_configuration
            and self._dp_spec.modbus_data_point_configuration.address
        ):
            self._address = (
                self._dp_spec.modbus_data_point_configuration.address
            )

        self._data_type: Optional[ModbusDataType] = None
        if (
            self._dp_spec.modbus_data_point_configuration
            and self._dp_spec.modbus_data_point_configuration.modbus_data_type
        ):
            self._data_type = (
                self._dp_spec.modbus_data_point_configuration.modbus_data_type
            )

        self._size = -1
        if (
            self._dp_spec.modbus_data_point_configuration
            and self._dp_spec.modbus_data_point_configuration.number_of_registers
        ):
            self._size = self._dp_spec.modbus_data_point_configuration.number_of_registers

        self._register_type: RegisterType = None
        if (
            self._dp_spec.modbus_data_point_configuration
            and self._dp_spec.modbus_data_point_configuration.register_type
        ):
            self._register_type = (
                self._dp_spec.modbus_data_point_configuration.register_type
            )

    async def set_val(self, value: Any):
        # convert to device units
        if (
            self._dp_spec.data_point.unit_conversion_multiplicator
            and self._dp_spec.data_point.unit_conversion_multiplicator != 1.0
        ):
            value = (
                float(value)
                / self._dp_spec.data_point.unit_conversion_multiplicator
            )

        # round to int if modbus type is int and DP type is not
        if is_float_type(
            self._dp_spec.data_point.data_type
        ) and not is_integer_type(
            self._dp_spec.modbus_data_point_configuration.modbus_data_type
        ):
            value = value_util.round_to_int(float(value))

        return await self._interface.write_data(
            self._register_type, self._address, self._data_type, value
        )

    async def get_val(self, skip_cache: bool = False) -> Any:
        # TODO implement skip_cache
        ret_value = await self._interface.read_data(
            self._register_type, self._address, self._size, self._data_type
        )

        # convert to DP units
        if (
            self._dp_spec.data_point.unit_conversion_multiplicator
            and self._dp_spec.data_point.unit_conversion_multiplicator != 1.0
        ):
            ret_value = (
                float(ret_value)
                * self._dp_spec.data_point.unit_conversion_multiplicator
            )

        # round to int if DP type is int and modbus type is not
        if is_integer_type(
            self._dp_spec.data_point.data_type
        ) and not is_float_type(
            self._dp_spec.modbus_data_point_configuration.modbus_data_type
        ):
            ret_value = value_util.round_to_int(float(ret_value))

        return ret_value

    def name(self) -> tuple[str, str]:
        return self._fp_name, self._dp_name

    def direction(self) -> DataDirectionProduct:
        if (
            self._dp_spec.data_point is None
            or self._dp_spec.data_point.data_direction is None
        ):
            raise Exception('missing data direction')
        return self._dp_spec.data_point.data_direction
    
    def unit(self) -> Units:
        if (
            self._dp_spec.data_point is None
            or self._dp_spec.data_point.unit is None
        ):
            return Units.NONE
        return self._dp_spec.data_point.unit


class ModbusFunctionalProfile(FunctionalProfile):
    """
    Implements a functional profile of a Modbus interface.
    """

    def __init__(
        self,
        fp_spec: ModbusFunctionalProfileSpec,
        interface: 'SGrModbusInterface',
    ):
        self._fp_spec = fp_spec
        self._interface = interface
        dps = [
            build_modbus_data_point(dp, self._fp_spec, self._interface)
            for dp in self._fp_spec.data_point_list.data_point_list_element
        ]
        self._data_points = {dp.name(): dp for dp in dps}

    def name(self) -> str:
        return self._fp_spec.functional_profile.functional_profile_name

    def get_data_points(self) -> dict[tuple[str, str], DataPoint]:
        return self._data_points


class SGrModbusInterface(SGrBaseInterface):
    """
    Implements a Modbus device interface.
    """

    def __init__(
        self,
        frame: DeviceFrame,
        sharedRTU: bool = False,
    ):
        self._inititalize_device(frame)
        if (
            self.device_frame.interface_list is None
            or self.device_frame.interface_list.modbus_interface is None
            or self.device_frame.interface_list.modbus_interface.modbus_interface_description
            is None
        ):
            raise Exception('Modbus interface is undefined')

        if self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_rtu:
            self.slave_id = get_rtu_slave_id(
                self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_rtu
            )
            self.serial_port = get_rtu_serial_port(
                self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_rtu
            )
            self.baudrate = get_rtu_baudrate(
                self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_rtu
            )
            self.parity = get_rtu_parity(
                self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_rtu
            )
        elif self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_tcp:
            self.slave_id = get_tcp_slave_id(
                self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_tcp
            )
            self.ip_address = get_tcp_address(
                self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_tcp
            )
            self.ip_port = get_tcp_port(
                self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_tcp
            )
        else:
            raise Exception('not Modbus RTU or TCP!')

        self.byte_order = get_endian(
            self.device_frame.interface_list.modbus_interface.modbus_interface_description
        )

        self.address_offset = get_address_offset(
            self.device_frame.interface_list.modbus_interface.modbus_interface_description
        )

        # build functional profiles
        fps = [
            ModbusFunctionalProfile(fp, self)
            for fp in self.device_frame.interface_list.modbus_interface.functional_profile_list.functional_profile_list_element
        ]
        self.functional_profiles = {fp.name(): fp for fp in fps}

        # unique string used in combination with shared Modbus client
        self._device_id = ''.join(random.choices(string.ascii_letters, k=8))
        self._client_wrapper: ModbusClientWrapper = None
        if (
            self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_interface_selection
            == ModbusInterfaceSelection.TCPIP
        ):
            self._client_wrapper = ModbusClientWrapper(
                '',
                SGrModbusTCPClient(
                    self.ip_address, self.ip_port, endianness=self.byte_order, addr_offset=self.address_offset
                ),
                shared=False,
            )
        elif (
            self.device_frame.interface_list.modbus_interface.modbus_interface_description.modbus_interface_selection
            == ModbusInterfaceSelection.RTU
        ):
            if sharedRTU:
                logger.debug('using shared RTU client')
                self._client_wrapper = register_shared_client(
                    self.serial_port,
                    self.parity,
                    self.baudrate,
                    device_id=self._device_id,
                )
            else:
                self._client_wrapper = ModbusClientWrapper(
                    '',
                    SGrModbusRTUClient(
                        self.serial_port,
                        self.parity,
                        self.baudrate,
                        endianness=self.byte_order,
                        addr_offset=self.address_offset
                    ),
                    shared=False,
                )
        else:
            raise Exception('Unsupported Modbus interface type')

    def __del__(self):
        if self._client_wrapper and self._client_wrapper.shared:
            unregister_shared_client(
                self.serial_port, device_id=self._device_id
            )

    def is_connected(self) -> bool:
        return self._client_wrapper.is_connected(self._device_id)

    async def connect_async(self):
        await self._client_wrapper.connect(self._device_id)

    async def disconnect_async(self):
        await self._client_wrapper.disconnect(self._device_id)

    async def read_data(
        self,
        reg_type: RegisterType,
        address: int,
        size: int,
        data_type: ModbusDataType,
    ) -> Any:
        """
        Reads data from the given Modbus address(es).
        """
        slave_id = self.slave_id
        if reg_type == RegisterType.INPUT_REGISTER:
            return await self._client_wrapper.client.read_input_registers(
                slave_id, address, size, data_type
            )
        elif reg_type == RegisterType.HOLD_REGISTER:
            return await self._client_wrapper.client.read_holding_registers(
                slave_id, address, size, data_type
            )
        elif reg_type == RegisterType.COIL:
            return await self._client_wrapper.client.read_coils(
                slave_id, address, size, data_type
            )
        elif reg_type == RegisterType.DISCRETE_INPUT:
            return await self._client_wrapper.client.read_discrete_inputs(
                slave_id, address, size, data_type
            )
        else:
            raise Exception(f'cannot read from register type {reg_type}')

    async def write_data(
        self,
        reg_type: RegisterType,
        address: int,
        data_type: ModbusDataType,
        value: Any,
    ) -> None:
        """
        Writes data to the given Modbus address(es).
        """
        slave_id = self.slave_id
        if reg_type == RegisterType.HOLD_REGISTER:
            await self._client_wrapper.client.write_holding_registers(
                slave_id, address, data_type, value
            )
        elif reg_type == RegisterType.COIL:
            await self._client_wrapper.client.write_coils(
                slave_id, address, data_type, value
            )
        else:
            raise Exception(f'cannot write to register type {reg_type}')

    def set_slave_id(self, slave_id: int):
        self.slave_id = slave_id
